import os
if os.path.exists('VERSION'):
    __VERSION__ = open('VERSION').read().strip()
else:
    __VERSION__ = "installed-but-no-version-file"

__all__ = ["TeleLib", "main_path", "telegram"]

import asyncio
import datetime
import json
import logging as lg
from typing import Optional, Tuple, Union
import coloredlogs
import httpx
from thread_py import ThreadPy
from telelib.tools.code_generator import CodeGenerator
from telelib.tools.scraper import Scraper
from telelib.telegram import getUpdates
import telelib.telegram as telegram

main_path = os.path.realpath(
    os.path.dirname(__file__) + '/../'
)


class Tools:
    documents_path = "https://core.telegram.org/bots/api"

    @staticmethod
    def ScraperRun():
        return Scraper().run("min.json")

    @staticmethod
    def CodeGeneratorRun():
        return CodeGenerator().run("min.json")


class Logger:
    _instance = None
    _ = None

    class LogLevel:
        DEBUG = lg.DEBUG
        INFO = lg.INFO
        WARNING = lg.WARNING
        ERROR = lg.ERROR
        CRITICAL = lg.CRITICAL
        FATAL = lg.FATAL
        WARN = lg.WARN
        NOTSET = lg.NOTSET

    format = "%(asctime)s [%(levelname)s] " \
        "[%(processName)s -> %(threadName)s] => | %(message)s"
    formatter = lg.Formatter(format)

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        self._ = self.get_logger(name="TeleLib")

    @classmethod
    def setup_logger(cls, name, log_file, level=LogLevel.INFO):

        handler = lg.FileHandler(log_file)
        handler.setFormatter(cls.formatter)

        logger = lg.getLogger(name)
        logger.setLevel(level)
        logger.addHandler(handler)

        return logger

    @classmethod
    def get_logger(
        cls,
        level: int = LogLevel.INFO,
        name="DEFAULT_LOGGER"
    ):
        if not os.path.exists("./logs/"):
            os.makedirs("./logs/")
            print(f"Created a directory at `./logs/`")

        if name is None:
            name = "NULL_LOGGER"

        today = datetime.date.today()

        lgr = cls.setup_logger(
            name=name,
            log_file=(
                f'./logs/{today.strftime("%Y-%m-%d")}-{name}.log'
            ),
            level=level
        )

        levels = {
            cls.LogLevel.DEBUG: "DEBUG",
            cls.LogLevel.INFO: "INFO",
            cls.LogLevel.WARNING: "WARNING",
            cls.LogLevel.ERROR: "ERROR",
            cls.LogLevel.CRITICAL: "CRITICAL",
            cls.LogLevel.FATAL: "FATAL",
            cls.LogLevel.WARN: "WARN",
            cls.LogLevel.NOTSET: "NOTSET",
        }

        coloredlogs.install(
            logger=lgr,
            fmt=cls.format,
            level=levels[level]
        )
        return lgr

    def lgr(self) -> lg.Logger:
        if self._ is None:
            self()

        return self._

    def __call__(self, *args, **kwds):
        return self._.info(msg=" ".join(args), **kwds)


class Telegram:
    class RequestData:
        def __init__(self, params, files={}):
            self.multipart_data = files
            self.json_parameters = params
            for key, value in self.json_parameters.copy().items():
                if value is None:
                    del self.json_parameters[key]
            for key, value in self.multipart_data.copy().items():
                if value is None:
                    del self.multipart_data[key]

        def __repr__(self) -> str:
            return f"{self.multipart_data}, {self.json_parameters}"

    USER_AGENT = "TelegramBotAPI/1.0"
    update_types = [
        'message',
        'edited_message',
        'channel_post',
        'edited_channel_post',
        'inline_query',
        'chosen_inline_result',
        'callback_query',
        'shipping_query',
        'pre_checkout_query',
        'poll',
        'poll_answer',
        'my_chat_member',
        'chat_member',
        'chat_join_request',
    ]

    _instance = None
    token = None
    api_url = None
    client_args = None
    client = None
    me = None

    def __repr__(self) -> str:
        return f"TelegramAPI(id={id(self)})"

    def __init__(self, token=None, api_url=None, proxy_url=None):
        Telegram.init(token, api_url, proxy_url)

    @classmethod
    def init(cls, token=None, api_url=None, proxy_url=None):
        if token is not None:
            Telegram.token = token
        if api_url is not None:
            Telegram.api_url = api_url
        timeout = httpx.Timeout(
            connect=10,
            read=10,
            write=10,
            pool=100,
        )
        limits = httpx.Limits(
            max_connections=100,
            max_keepalive_connections=100,
        )
        Telegram.client_args = {
            "timeout": timeout,
            "proxies": proxy_url,
            "limits": limits,
        }
        Telegram.client = httpx.AsyncClient(**cls.client_args)

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    @classmethod
    async def construct(cls):
        cls.client = httpx.AsyncClient(**cls.client_args)
        return cls.client

    @classmethod
    async def destruct(cls):
        if cls.client.is_closed:
            raise Exception("Connection is already closed.")

        await cls.client.aclose()

    async def __aenter__(self):
        await self.construct()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.destruct()

    @classmethod
    async def __req(
        cls,
        url: str,
        method: str,
        request_data: Optional[RequestData] = None,
    ) -> Tuple[int, bytes]:
        if cls.client is None:
            cls.client = await cls.construct()

        if cls.client.is_closed:
            raise RuntimeError("Client is Closed.")

        files = request_data.multipart_data if request_data else None
        data = request_data.json_parameters if request_data else None

        try:
            res = await cls.client.request(
                method=method,
                url=url,
                headers={"User-Agent": cls.USER_AGENT},
                timeout=cls.client_args['timeout'],
                files=files,
                json=data
            )
        except httpx.TimeoutException as err:
            if isinstance(err, httpx.PoolTimeout):
                raise Exception("Pool is full")
            raise err
        except httpx.HTTPError as err:
            raise Exception(f"httpx HTTPError: {err}")

        return res.status_code, (json.loads(res.content))

    @classmethod
    async def is_ok(cls, res: Tuple[int, dict]) -> bool:
        if res[0] == 200:
            if res[1]["ok"]:
                return res[1]['result']
        return False

    @classmethod
    async def __method(
        cls,
        method: str,
        request_data: Optional[Union[RequestData, dict]] = None,
        **kwargs
    ) -> dict:

        url = f"{cls.api_url}{cls.token}/{method}"
        _method = "GET" if request_data is None else "POST"
        if request_data is None:
            request_data = cls.RequestData(kwargs)

        if isinstance(request_data, cls.RequestData):
            res = await cls.__req(
                url,
                _method,
                request_data
            )
        else:
            res = await cls.__req(
                url,
                _method,
                cls.RequestData(request_data)
            )
        return res

    @classmethod
    async def call_method(
        cls,
        method: str,
        **kwargs
    ) -> dict:
        return await cls.is_ok(
            await cls.__method(
                method,
                **kwargs
            )
        )

    @classmethod
    async def __call__(cls, method: "telegram.Types.DefaultMethod") -> dict:
        _method, args = method._call()
        method._called = True
        method._res = await cls.call_method(_method, **args)
        return method


class TeleLib:
    Logger = Logger
    Tools = Tools
    tg = None

    _thread = None
    token = None
    api_url = "https://api.telegram.org/bot"
    _event_loop = None
    _instance = None

    debug = False
    logger = None
    tg = None
    is_running = True

    def __repr__(self) -> str:
        return (f"TeleLib v{__VERSION__}")

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, token=None, api_url=None, debug=True):
        if token is not None:
            TeleLib.token = token

        if api_url is not None:
            TeleLib.api_url = api_url

        TeleLib._thread = ThreadPy()
        TeleLib.debug = debug

        TeleLib._event_loop = asyncio.new_event_loop()
        TeleLib._event_loop.set_debug(TeleLib.debug)
        TeleLib.logger = self.Logger()
        TeleLib.tg = Telegram(TeleLib.token, TeleLib.api_url)

    @property
    def event_loop(self):
        if self._event_loop is None:
            TeleLib._event_loop = asyncio.new_event_loop()
        return self._event_loop

    @classmethod
    def AsyncRun(cls, func, *args, **kwargs):
        async def runner():
            await func(*args, **kwargs)

        return asyncio.run_coroutine_threadsafe(
            runner(),
            cls._event_loop
        )

    class Bot:
        event_handlers = {}

        @classmethod
        def set_event_handler(cls, event, func):
            cls.event_handlers[event] = func

        @classmethod
        async def start(cls, offset=0):
            TeleLib.logger("Starting TeleLib Bot Instance")

            while TeleLib.is_running:
                await asyncio.sleep(0.01)
                await TeleLib.tg(updates := getUpdates(offset=offset))

                def _(update):
                    for event, event_func in cls.event_handlers.items():
                        for update_type in Telegram.update_types:
                            if type(getattr(update, update_type)) == event:
                                try:
                                    ThreadPy.create_and_start_thread(
                                        name=f"{update.update_id}-Thread",
                                        target=TeleLib.AsyncRun,
                                        args=(
                                            event_func,
                                            getattr(update, update_type)
                                        )
                                    )
                                except Exception as err:
                                    print(err)

                                return True

                for update in updates.result():
                    offset = update.update_id + 1
                    _(update)

                    if 'main' in cls.event_handlers:
                        ThreadPy.create_and_start_thread(
                            name=f"{update.update_id}-Thread",
                            target=TeleLib.AsyncRun,
                            args=(
                                cls.event_handlers['main'],
                                update
                            )
                        )

    # Extra Python Tools

    def start(self):
        TeleLib.logger("Starting TeleLib")
        self.event_loop.run_until_complete(TeleLib.Bot.start())



import setuptools

setuptools.setup(
    name="telelib",
    description="A Telegram Wrapper Written in ?",
    author="Mohammad Mahdi Afshar",
    author_email="me@mamad.dev",
    version=open("VERSION").read().strip(),
    license="The Unlicensed",
    packages=setuptools.find_packages(),
    dependencies=[
        "thread_py",
        "xhttp",
    ],
    classifiers=[
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Communications :: Chat",
        "Topic :: Internet",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires=">=3.8",
)



import json
import telelib
import re
import string

import requests
from bs4 import BeautifulSoup
from bs4.element import Tag


# i used  https://github.com/PaulSonOfLars/telegram-bot-api-spec
# as a base for this scraper
class Scraper:
    def __repr__(self) -> str:
        return f"Telegram Document Scraper v{telelib.__VERSION__}"

    def __init__(self):
        self.METHODS = "methods"
        self.TYPES = "types"
        self.TG_CORE_TYPES = ["String", "Boolean", "Integer", "Float"]

        self.APPROVED_NO_SUBTYPES = {"InputFile"}
        self.APPROVED_MULTI_RETURNS = [["Message", "Boolean"]]

    def __get_fields(
        self,
        curr_name: str,
        curr_type: str,
        x: Tag,
        items: dict,
        url: str
    ):
        body = x.find("tbody")
        fields = []
        for tr in body.find_all("tr"):
            children = list(tr.find_all("td"))
            if curr_type == self.TYPES and len(children) == 3:
                desc = self.__clean_tg_field_description(children[2], url)
                fields.append(
                    {
                        "name": children[0].get_text(),
                        "types": self.__clean_tg_type(children[1].get_text()),
                        "required": not desc.startswith("Optional. "),
                        "description": desc,
                    }
                )

            elif curr_type == self.METHODS and len(children) == 4:
                fields.append(
                    {
                        "name": children[0].get_text(),
                        "types": self.__clean_tg_type(children[1].get_text()),
                        "required": children[2].get_text() == "Yes",
                        "description": self.__clean_tg_field_description(
                            children[3], url
                        ),
                    }
                )

            else:
                print("An unexpected state has occurred!")
                print("Type:", curr_type)
                print("Name:", curr_name)
                print("Number of children:", len(children))
                print(children)
                exit(1)

        items[curr_type][curr_name]["fields"] = fields

    def __get_method_return_type(
        self,
        curr_name: str,
        curr_type: str,
        description_items: list[str],
        items: dict,
    ):
        description = "\n".join(description_items)
        ret_search = re.search(
            "(?:on success,|returns)([^.]*)(?:on success)?",
            description,
            re.IGNORECASE,
        )
        ret_search2 = re.search(
            "([^.]*)(?:is returned)", description, re.IGNORECASE
        )
        if ret_search:
            self.__extract_return_type(
                curr_type, curr_name, ret_search.group(1).strip(), items
            )
        elif ret_search2:
            self.__extract_return_type(
                curr_type, curr_name, ret_search2.group(1).strip(), items
            )

    def __extract_return_type(
        self,
        curr_type: str,
        curr_name: str,
        ret_str: str,
        items: dict
    ):
        array_match = re.search(r"(?:array of )+(\w*)", ret_str, re.IGNORECASE)
        if array_match:
            ret = self.__clean_tg_type(array_match.group(1))
            rets = [f"Array of {r}" for r in ret]
            items[curr_type][curr_name]["returns"] = rets
        else:
            words = ret_str.split()
            rets = [
                r
                for ret in words
                for r in self.__clean_tg_type(
                    ret.translate(str.maketrans("", "", string.punctuation))
                )
                if ret[0].isupper()
            ]
            items[curr_type][curr_name]["returns"] = rets

    def __clean_tg_field_description(
        self,
        t: Tag,
        url: str
    ) -> str:
        return " ".join(self.__clean_tg_description(t, url))

    def __clean_tg_description(
        self,
        t: Tag,
        url: str
    ) -> list[str]:
        for i in t.find_all("img"):
            i.replace_with(i.get("alt"))

        for br in t.find_all("br"):
            br.replace_with("\n")

        for a in t.find_all("a"):
            anchor_text = a.get_text()
            if "»" not in anchor_text:
                continue

            link = a.get("href")

            if link.startswith("#"):
                link = url + link

            elif link.startswith("/"):
                link = telelib.TeleLib.Tools.documents_path + link

            anchor_text = anchor_text.replace(" »", ": " + link)
            a.replace_with(anchor_text)

        text = t.get_text()

        text = re.sub(r"(\s){2,}", r"\1", text)
        text = text.replace("”", '"').replace("“", '"')
        text = text.replace("…", "...")
        text = text.replace("\u2013", "-")
        text = text.replace("\u2014", "-")
        text = text.replace("\u2019", "'")
        return [t.strip() for t in text.split("\n") if t.strip()]

    def __get_proper_type(
        self,
        t: str
    ) -> str:
        if t == "Messages":
            return "Message"

        elif t == "Float number":
            return "Float"

        elif t == "Int":
            return "Integer"

        elif t == "True" or t == "Bool":
            return "Boolean"

        return t

    def __clean_tg_type(
        self,
        t: str
    ) -> list[str]:
        pref = ""
        if t.startswith("Array of "):
            pref = "Array of "
            t = t[len("Array of "):]

        fixed_ors = [x.strip() for x in t.split(" or ")]
        fixed_ands = [x.strip() for fo in fixed_ors for x in fo.split(" and ")]
        fixed_commas = [x.strip() for fa in fixed_ands for x in fa.split(", ")]
        return [pref + self.__get_proper_type(x) for x in fixed_commas]

    def __get_subtypes(
        self, curr_name: str, curr_type: str, x: Tag, items: dict, url: str
    ):
        if curr_name == "InputFile":
            return

        list_contents = []
        for li in x.find_all("li"):
            list_contents.extend(self.__clean_tg_description(li, url))

        if curr_type == self.TYPES:
            items[curr_type][curr_name]["subtypes"] = list_contents

        items[curr_type][curr_name]["description"] += [
            f"- {s}" for s in list_contents
        ]

    def __get(
        self,
        url: str
    ) -> dict:
        r = requests.get(url)
        soup = BeautifulSoup(r.text, features="html5lib")
        dev_rules = soup.find("div", {"id": "dev_page_content"})
        curr_type = ""
        curr_name = ""

        items = {
            self.METHODS: dict(),
            self.TYPES: dict(),
        }

        for x in list(dev_rules.children):
            if x.name == "h3" or x.name == "hr":
                curr_name = ""
                curr_type = ""

            if x.name == "h4":
                anchor = x.find("a")
                name = anchor.get("name")
                if name and "-" in name:
                    curr_name = ""
                    curr_type = ""
                    continue

                curr_name, curr_type = self.__type_name(x, anchor, items, url)

            if not curr_type or not curr_name:
                continue

            if x.name == "p":
                items[curr_type][curr_name].setdefault(
                    "description", []
                ).extend(self.__clean_tg_description(x, url))

            if x.name == "table":
                self.__get_fields(curr_name, curr_type, x, items, url)

            if x.name == "ul":
                self.__get_subtypes(curr_name, curr_type, x, items, url)

            if curr_type == self.METHODS and items[curr_type][curr_name].get(
                "description"
            ):
                self.__get_method_return_type(
                    curr_name,
                    curr_type,
                    items[curr_type][curr_name].get("description"),
                    items,
                )

        return items

    def __type_name(
        self,
        t: Tag,
        anchor: Tag,
        items: dict,
        url: str
    ):
        if t.text[0].isupper():
            curr_type = self.TYPES
        else:
            curr_type = self.METHODS
        curr_name = t.get_text()
        items[curr_type][curr_name] = {"name": curr_name}

        href = anchor.get("href")
        if href:
            items[curr_type][curr_name]["href"] = url + href

        return curr_name, curr_type

    def __verify_types(
        self,
        items: dict
    ) -> bool:
        issue_found = False

        for type_name, values in items[self.TYPES].items():
            # check all values have a URL
            if not values.get("href"):
                print(f"{type_name} has no link!")
                issue_found = True
                continue

            fields = values.get("fields", [])
            if len(fields) == 0:
                subtypes = values.get("subtypes", [])
                description = "".join(values.get("description", [])).lower()
                # Some types are abstract and have no information
                # or subtypes - this is mentioned in their description.
                # Otherwise, check if they're manually approved.
                if not subtypes and not (
                    "currently holds no information" in description
                    or type_name in self.APPROVED_NO_SUBTYPES
                ):
                    print(
                        "TYPE",
                        type_name,
                        "has no fields or subtypes, and is not approved",
                    )
                    issue_found = True
                    continue

                for st in subtypes:
                    if st in items[self.TYPES]:
                        items[self.TYPES][st].setdefault(
                            "subtype_of", []
                        ).append(type_name)
                    else:
                        print("TYPE", type_name, "USES INVALID SUBTYPE", st)
                        issue_found = True

            # check all parameter types are valid
            for param in fields:
                field_types = param.get("types")
                for field_type_name in field_types:
                    while field_type_name.startswith("Array of "):
                        field_type_name = field_type_name[len("Array of "):]

                    if (
                        field_type_name not in items[self.TYPES]
                        and field_type_name not in self.TG_CORE_TYPES
                    ):
                        print("UNKNOWN FIELD TYPE", field_type_name)
                        issue_found = True

        return issue_found

    def __verify_methods(
        self,
        items: dict
    ) -> bool:
        issue_found = False
        # Type check all methods
        for method, values in items[self.METHODS].items():
            # check all values have a URL
            if not values.get("href"):
                print(f"{method} has no link!")
                issue_found = True
                continue

            # check all methods have a return
            returns = values.get("returns")
            if not returns:
                print(f"{method} has no return types!")
                issue_found = True
                continue

            # If we have multiple returns, this is an edge case,
            # but not a deal-breaker; check that output.
            # Some multi-returns are common and pre-approved.
            if len(returns) > 1 and returns not in self.APPROVED_MULTI_RETURNS:
                print(f"{method} has multiple return types: {returns}")

            # check all parameter types are valid
            for param in values.get("fields", []):
                types = param.get("types")
                for t in types:
                    while t.startswith("Array of "):
                        t = t[len("Array of "):]

                    if (
                        t not in items[self.TYPES]
                        and t not in self.TG_CORE_TYPES
                    ):
                        issue_found = True
                        print("UNKNOWN PARAM TYPE", t)

            # check all return types are valid
            for ret in values.get("returns", []):
                while ret.startswith("Array of "):
                    ret = ret[len("Array of "):]

                if (
                    ret not in items[self.TYPES]
                    and ret not in self.TG_CORE_TYPES
                ):
                    issue_found = True
                    print("UNKNOWN RETURN TYPE", ret)

        return issue_found

    def run(
        self,
        filename,
        indent=None
    ):
        items = self.__get(telelib.TeleLib.Tools.documents_path)
        if self.__verify_types(items) or self.__verify_methods(items):
            raise RuntimeError("Failed to validate schema.")

        file_name = telelib.main_path + "/generated/" + filename

        with open(file_name, "w") as f:
            json.dump(items, f, indent=indent)

        return "Successfully generated json Scheme at " + file_name
